#' Calculate the generalized AC1 coefficient.
#'
#' Estimate, calculate confidence intervals, and perform hypothesis tests
#'    for the generalized AC1 coefficient.
#'
#' The confidence interval and hypothesis test is calculated using asymptotic theory
#'    for the `plugin` estimator under the `"bp"` model; otherwise the
#'    bias-corrected and accelerated boostrap is used. For the confidence interval
#'    construction of Gwet (2008), see `[irrCAC::gwet.ac1.dist]`.
#'
#' RIGHT NOW ONLY PLUGIN WITH `"bp"` IS IMPLEMENTED!
#'
#' The output is modelled after the tests in the `stats` package, e.g.
#'    `stats::prop.test` and `stats::t.test`.
#'
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge. (CHANGE TO KRIPPENDORFF FORM.)
#' @param alternative a character string specifying the alternative hypothesis,
#'    must be one of `"two.sided"` (default), `"greater"` or `"less"`. Used both
#'    for the test and the confidence interval.
#' @param conf_level Confidence level for the confidence interval. Defaults to
#'    to `0.95`.
#' @param model The model used in the estimation of the coefficient. The
#'    `"general"` model makes no additional assumptions. The `"cxf"` (Cohen-x-Fleiss)
#'    model assumes the true distribution equals the marginal distribution.
#'    The `"equality"` model assumes all guessing distributions are equal.
#'    `"bp"`model assumes that the guessing distribution is uniform, the
#'    `"fleiss"` model assumes that the guessing distribution equals the
#'    marginal distribution of ratings.
#' @param method The estimation method. Either `"plugin"`, `"ml"`, or `"gwet"`.
#'    Defaults to `"plugin"`, which works both for the Fleiss and Gwet models.
#'    The option `"gwet"` uses the approximation method of Gwet (2008), and is
#'    not recommended.
#' @param gamma The AC1 coefficient of the null-hypothesis. Defaults to `0`,
#'    which is the minimal value the AC1 coefficient.
#' @param skip If `TRUE`, skips boostrapping and only calcutes the AC1
#'    coefficient. Defaults to `FALSE`
#' @param cat Number of categories. Defaults the number of unique values in the
#'    data.
#' @return A list with class `"htest"` containing the following components:
#'   \describe{
#'     \item{`statistic`}{the value of the test statistic.}
#'     \item{`p.value`}{the p-value for the test.}
#'     \item{`conf.int`}{a confidence interval for the mean appropriate to the
#'     specified alternative hypothesis.}
#'     \item{`estimate`}{the estimated AC1 coefficient.}
#'     \item{`null.value`}{the specified hypothesized value of the AC1
#'     coefficient.}
#'     \item{`stderrr`}{the standard error of the estimator.}
#'     \item{`alternative`}{a character string describing the alternative
#'     hypothesis.}
#'     \item{`method`}{a character string indicating if `"plugin"` or `"bp"`
#'     was used and the model used (either  `"bp"` or `"fleiss"`.)}
#'     \item{`data.name`}{a character string giving the name of the data.}
#'   }
#' @export
#' @references
#' Efron, B. (1987). Better Bootstrap Confidence Intervals. Journal of the
#'     American Statistical Association, 82(397), 171–185.
#'     https://doi.org/10.2307/2289144
#'
#' Gwet, K. L. (2014). Handbook of Inter-Rater Reliability.
#'    Advanced Analytics, LLC.
#'
#' Gwet, K. L. (2008). Computing inter-rater reliability and its variance in
#'    the presence of high agreement. The British Journal of Mathematical and
#'    Statistical Psychology, 61, 29–48.
#'    https://doi.org/10.1348/000711006X126600
#'
#' van Oest, R. (2019). A new coefficient of interrater agreement: The
#'    challenge of highly unequal category proportions. Psychological Methods,
#'    24(4), 439–451. https://doi.org/10.1037/met0000183
#' @examples
#'   # Using the default plugin method.
#'   ac1(triplet_to_raw(dat.zapf2016))
#'
#'   # Using Gwet's approximation method.
#'   # ac1(triplet_to_raw(dat.zapf2016), method = "gwet")

ac1 = function(x, alternative = c("two.sided", "greater", "less"),
               conf_level = 0.95, method = c("plugin", "empirical", "gwet"),
               model = c("bp", "fleiss"),
               gamma = 0, skip = FALSE, cat = length(unique(c(x)))) {

  d_name = deparse1(substitute(x))
  method = match.arg(method)
  model = match.arg(model)
  alternative = match.arg(alternative)

  if(model == "fleiss" & method == "gwet")
    stop('The "gwet" method for estimation of the AC1 coefficient is only defined for the "bp" model, not the "fleiss" model.')

  pi_hat = table(x) / length(x)
  n = nrow(x)

  pa = pwa_hat(x)

  # Estimate.
  gamma_hat = (cat * pa - 1) / (cat - 2 + pa)

  if (skip) return(gamma_hat)

  if(method == "empirical") {

    stats::sd(mua_hats(x))  * (cat - 1)^2 / (cat + pa - 2)^2

  } else if (method == "plugin") {

    p2a_hat = function(x) {
      map = utils::combn(ncol(x), 2)
      mean(apply(
        X = apply(arrangements::permutations(choose(ncol(x), 2), 2, replace = TRUE), 1, function(i) map[, i]),
        MARGIN = 2,
        FUN = function(indices) mean(x[, indices[1]] == x[, indices[2]] &
                                       x[, indices[3]] == x[, indices[4]])
      ))
    }

    p2a = p2a_hat(x)

    if(p2a < pa ^ 2) {
      tau = 0.3 * (cat - 1)^2 / (cat + pa - 2)^2
    } else {
      tau = sqrt(p2a - pa ^ 2) * (cat - 1)^2 / (cat + pa - 2)^2
    }

  }



  # Make confidence interval
  multiplier = c(stats::qt((1 - conf_level)/2, n - 1), -stats::qt((1 - conf_level)/2, n - 1))
  confidence_interval = gamma_hat + tau / sqrt(n - 1) * multiplier
  attr(confidence_interval, "conf.level") = conf_level

  # Make statistic in p-value.
  statistic = sqrt(n) / tau * (gamma_hat - gamma)
  p_value = stats::pnorm(-abs(statistic)) / 2

  return_value = list(
    statistic = c(t = statistic),
    p.value = p_value,
    estimate = c("AC1 (gamma)" = gamma_hat),
    null.value = c(AC1 = gamma),
    conf.int = confidence_interval,
    stderr = tau / sqrt(n - 1),
    alternative = alternative,
    method = paste0("AC1 using the '", method, "' method and the ", model, " model."),
    data.name = d_name
  )

  class(return_value) = "htest"

  return_value

}

