#' Calculate the sample agreement coefficient
#'
#' The sample agreement coefficient equals the sample Cohen's kappa when there
#'    are two raters and no missing values. It equals Krippendorff's [alpha] when
#'    every judge rates one item and every item is judge exactly twice. It
#'    equals the sample [pi] when the marginals for all the judges are the same.
#'
#' `agreement` takes its input data on triplet form. This is a data frame with
#'    three columns: `rating`, `item`, `judge`. Repeated item-judge pairs are
#'    currently not supported. That is, every judge rates every item at most
#'    once. You may convert raw data to triplet form data using the function
#'    [transform_data].
#'
#' @param x Agreement data on triplet form; see the details.
#' @param loss The loss function to use. Must be symmetric.
#' @return The sample agreement coefficient.
#' @seealso The sample [pi] and Krippendorff's [alpha].
#' @references
#' Cohen, J. (1968). Weighted kappa: Nominal scale agreement with provision for
#' scaled disagreement or partial credit. Psychological Bulletin, 70(4),
#' 213–220. https://doi.org/10.1037/h0026256
#'
#' Krippendorff, K. (2018). Content analysis: An introduction to its methodology
#' (4th ed.). SAGE Publications.
#'
#' Scott, W. A. (1955). Reliability of Content Analysis: The Case of Nominal
#' Scale Coding. Public Opinion Quarterly, 19(3), 321–325.
#' https://doi.org/10.1086/266577
#'
#' Fleiss, J. L. (1971). Measuring nominal scale agreement among many raters.
#' Psychological Bulletin, 76(5), 378–382. https://doi.org/10.1037/h0031619
#' @export
agreement = function(x, loss = function(x, y) x != y) {

  m_i = table(x[, 2])
  n_o = sum(m_i * (m_i - 1))

  k_j = table(x[, 3])
  k_j_matrix = tcrossprod(k_j, k_j)
  diag(k_j_matrix) = 0
  n_c = sum(k_j_matrix)

  item_loss = function(y) {
    indices = arrangements::combinations(seq(length(y)), 2)
    sum(apply(indices, 1, function(index) loss(y[index[1]], y[index[2]])))
  }

  x = x[order(tibble::deframe(x[, 2])), ]
  y = x$rating
  upper = 0
  left_index = 1

  for(i in m_i) {
    right_index = left_index + i - 1
    upper = upper + item_loss(y[left_index:right_index])
    left_index = right_index + 1
  }

  upper = 2 * upper

  n = nrow(x)
  judge_loss = function(y, left_index, right_index) {

    numbers = seq(right_index + 1, n)
    sum(sapply(seq(left_index, right_index), function(i) {
      sum(sapply(numbers, function(index) loss(y[i], y[index])))
    }))

  }

  x = x[order(tibble::deframe(x[, 3])), ]
  y = x$rating
  lower = 0
  left_index = 1

  for(i in k_j[-length(k_j)]) {
    right_index = left_index + i - 1
    lower = lower + judge_loss(y, left_index, right_index)
    left_index = right_index + 1
  }

  lower = 2 * lower

  1 - (upper / n_o) / (lower / n_c)


}

#' Calculate the sample pi and Krippendorff alpha
#'
#' The sample pi subsumes Scott's pi, defined for two raters, and Fleiss' kappa,
#'    defined for `m` raters. Krippendorff's alpha is a simple transformation of
#'    the sample pi, \eqn{\alpha = \pi + 1/n * (1 - \pi)}, where \eqn{n} is the
#'    number of rows in `x`.
#'
#' Under the quadratic loss, the pi coefficeint is the Pearson correlation
#'    between paired data. Krippendorff's [alpha] equals the sample
#'    [agreement] when every judge rates one item and every item is
#'    judge exactly twice. The sample [pi] equals the sample [agreement]
#'    coefficient when the marginals for all the judges are the same. In
#'    particular, the sample [pi] equals Cohen's kappa when there are two
#'    raters with no missing values and the same marginals.
#'
#' @param x Agreement data on triplet form.
#' @param loss The loss function to use. Must be symmetric.
#' @return The sample pi coefficient or Krippendorff alpha.
#' @seealso The sample [agreement] coefficient.
#' @references
#' Cohen, J. (1968). Weighted kappa: Nominal scale agreement with provision for
#' scaled disagreement or partial credit. Psychological Bulletin, 70(4),
#' 213–220. https://doi.org/10.1037/h0026256
#'
#' Krippendorff, K. (2018). Content analysis: An introduction to its methodology
#' (4th ed.). SAGE Publications.
#'
#' Scott, W. A. (1955). Reliability of Content Analysis: The Case of Nominal
#' Scale Coding. Public Opinion Quarterly, 19(3), 321–325.
#' https://doi.org/10.1086/266577
#'
#' Fleiss, J. L. (1971). Measuring nominal scale agreement among many raters.
#' Psychological Bulletin, 76(5), 378–382. https://doi.org/10.1037/h0031619
#' @export
pi = function(x, loss = function(x, y) x != y) {

  m_i = table(x[, 2])
  n_o = sum(m_i * (m_i - 1)) / 2
  n = nrow(x)
  n_c = n ^ 2 / 2

  item_loss = function(y) {
    indices = arrangements::combinations(seq(length(y)), 2)
    sum(apply(indices, 1, function(index) loss(y[index[1]], y[index[2]])))
  }

  x = x[order(tibble::deframe(x[, 2])), ]
  y = x$rating
  upper = 0
  left_index = 1

  for(i in m_i) {
    right_index = left_index + i - 1
    upper = upper + item_loss(y[left_index:right_index])
    left_index = right_index + 1
  }

  lower = 0
  for(i in seq(from = 1, to = n - 1)) {
    for(j in seq(from = i + 1, to = n)) {
      lower = lower + loss(y[i], y[j])
    }
  }


  1 - (upper / n_o) / (lower / n_c)

}

#' @rdname pi
#' @export
alpha = function(x, loss = function(x, y) x != y) {
  pi_hat = pi(x, loss)
  pi_hat + 1/nrow(x) * (1 - pi_hat)
}
