#' Estimate a multi-rater weighted kappa under a rectangular design.
#'
#' @param x Agreement data on the wide form, ratings as rows and raters as
#'    raters as columns. See `dat.zapf2016` for an example on data on the wide
#'    form. If `type = "fleiss"`, the data can be on Fleiss form instead, where
#'    the rating options are columns, but the `fleiss_for` must be set to true.
#'    See `dat.fleiss1971` for an example on data on the Fleiss form.
#' @param disagreement Choose from `"nominal"`, `"absolute"`, `"quadratic"`,
#'    or `"hubert"`.
#' @param type The type of disagreement measure used. Defaults to `cohen`.
#' @param g The arity of the disagreement function. Defaults to `ncol(x)` if
#'   `disagreement = "hubert"`. Defaults to `2` for any other disagreement.
#' @param alternative a character string specifying the alternative hypothesis,
#'    must be one of `"two.sided"` (default), `"greater"` or `"less"`. Used both
#'    for the test and the confidence interval.
#' @param conf_level Confidence level for the confidence interval. Defaults to
#'    to `0.95`.
#' @param method Method for confidence interval construction. Defaults to the
#'    `arcsine` method, which performs slightly better than the others.
#' @param skip If `TRUE`, skips inference and calculates the kappa coefficient
#'    only. Defaults to `FALSE`
#' @param fleiss_form If `TRUE`, the data is on Fleiss form.
#' @param kappa The kappa of the null-hypothesis. Defaults to `0`.
#' @return The weighted multi-rater kappa coefficient
#' @examples
#' library("agreer")
#'
#' # The Zapf et al. (2016) data is on wide form, and can be used to calculate
#' # both Cohen's kappa and Fleiss' kappa.
#' kappa(dat.zapf2016, "quadratic", "cohen", g = 3, method = "basic")
#' kappa(dat.zapf2016, "hubert", "cohen", g = 4, method = "fisher")
#' kappa(dat.zapf2016, "absolute", "cohen", g = 2, method = "arcsine")
#'
#' # The Fleiss (1971) data, on the other hand, is on Fleiss form, and can only
#' # be used for Fleiss' kappa.
#' kappa(dat.fleiss1971, "nominal", "fleiss", g = 2, method = "basic")
#' kappa(dat.fleiss1971, "hubert", "fleiss", g = 4, method = "fisher")
#' kappa(dat.fleiss1971, "absolute", "fleiss", g = 3, method = "arcsine")
#' @export

kappa <- function(x,
                  disagreement = c("nominal", "absolute", "quadratic", "hubert"),
                  type = c("cohen", "fleiss"),
                  g = NULL,
                  alternative = c("two.sided", "greater", "less"),
                  conf_level = 0.95,
                  method = c("arcsine", "fisher", "basic"),
                  kappa = 0,
                  skip = FALSE,
                  fleiss_form = FALSE) {
  d_name <- deparse1(substitute(x))
  disagreement <- match.arg(disagreement)
  type <- match.arg(type)
  method <- match.arg(method)
  alternative <- match.arg(alternative)
  x <- prepare_data(x, fleiss_form, type)
  g <- prepare_g(g, disagreement, ncol(x))

  if (stats::var(c(x)) != 0) {

    hats <- kappa_calc(x, disagreement = disagreement, type = type, g = g)
    kappa_hat <- hats[1]
    sigma_hat <- sqrt(max(hats[2], 10e-06))
    ci <- ci(kappa_hat, sigma_hat, nrow(x), method, conf_level, alternative)
    statistic <- unname(sqrt(nrow(x)) / sigma_hat * (kappa_hat - kappa))
    p_value <- p_value(statistic, nrow(x), alternative)

  } else {
    kappa_hat <- 1
    sigma_hat <- p_value <- 0
    ci <- c(0, 1)
    attr(ci, "conf.level") <- conf_level
    attr(ci, "alternative") <- alternative
    attr(ci, "method") <- method
    statistic <- 0
    message("        All ratings are equal!
* The estimated kappa is 1 for any disagreent function and chance
  disagreement type.
* Cannot calculate the standard deviation, which is set to 0.
* The confidence interval and p-values are arbitrary.")
  }

  construct_htest(
    kappa_hat, type, g, disagreement, method, statistic,
    p_value, sigma_hat, kappa, ci, nrow(x), alternative,
    d_name, fleiss_form, ncol(x)
  )
}
