#' Estimate agreement coefficients under a rectangular design.
#'
#' We strongly recommend Cohen's kappa.
#' Cohen's kappa
#'
#' Fleiss' kappa is calculatated under the assumption that
#'
#' The weighted Brennan--Prediger coefficient
#'
#' Krippendorff's alpha is equals Fleiss' kappa plus a small correction term.
#'
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge.
#' @param alternative a character string specifying the alternative hypothesis,
#'    must be one of `"two.sided"` (default), `"greater"` or `"less"`. Used both
#'    for the test and the confidence interval.
#' @param conf_level Confidence level for the confidence interval. Defaults to
#'    to `0.95`.
#' @param weights Choose from `"quadratic"`, `"unit"`, `"abs"`, or supply your
#'    own function.
#' @param method Method for confidence interval construction.
#' @param kappa The kappa of the null-hypothesis. Defaults to `0`.
#' @param skip If `TRUE`, skips inference and calculates the kappa coefficient
#'    only. Defaults to `FALSE`
#' @param count If `TRUE`, uses a counting algorithm. This is faster for
#'    categorical data, but slower for continuous data.. Defaults to `NULL`,
#'    which evaluates to `TRUE` if the number of categories in the data
#'    is less than `sqrt(n)`, where `n` is the number of rows.
#' @return The weighted multirater kappa coefficient
#' @export
cohen_kappa = function(
  x,
  alternative = c("two.sided", "greater", "less"),
  conf_level = 0.95,
  weights = "unit",
  kappa = 0,
  method = c("arcsine", "fisher", "basic"),
  skip = FALSE,
  count = NULL) {

  d_name = deparse1(substitute(x))
  w_name = deparse1(substitute(weights))
  method = match.arg(method)

  if(skip) {

    estimate = function(x) {
      pwc = pwc_hat(x, w = weights, count = count)
      pwa = pwa_hat(x, w = weights)
      (pwa - pwc) / (1 - pwc)
    }

    return(estimate(x))

  }

  alternative = match.arg(alternative)

  n = nrow(x)
  hats = var_kw(x, w = weights, type = "cohen", count = count)
  sigma_hat = sqrt(hats$sigma_hat)
  kappa_hat = hats$kappa_hat

  # Make confidence interval
  if(alternative == "two.sided") {
    multiplier = c(stats::qt((1 - conf_level)/2, n - 1), -stats::qt((1 - conf_level)/2, n - 1))
  } else if (alternative == "greater") {
    multiplier = c(stats::qt((1 - conf_level), n - 1), Inf)
  } else if (alternative == "less") {
    multiplier = c(-Inf, -stats::qt((1 - conf_level), n - 1))
  }

  if(method == "basic") {
    confidence_interval = kappa_hat + sigma_hat / sqrt(n - 1) * multiplier
    confidence_interval[2] = min(1, confidence_interval[2])
  } else if (method == "arcsine") {
    std = sigma_hat / sqrt(1 - kappa_hat^2)
    confidence_interval = sin(asin(kappa_hat) + std / sqrt(n - 1) * multiplier)
  } else if (method == "fisher") {
    std = sigma_hat / (1 - kappa_hat^2)
    confidence_interval = tanh(atanh(kappa_hat) + std / sqrt(n - 1) * multiplier)
  }

  confidence_interval[2] = min(1, confidence_interval[2])
  attr(confidence_interval, "conf.level") = conf_level
  attr(confidence_interval, "alternative") = alternative
  attr(confidence_interval, "method") = method

  # Make statistic in p-value.
  statistic = sqrt(n) / sigma_hat * (kappa_hat - kappa)
  p_value = stats::pnorm(-abs(statistic)) / 2


  return_value = list(
    statistic = c(t = statistic),
    p.value = p_value,
    estimate = c("Cohen's kappa" = kappa_hat),
    null.value = c(kappa = kappa),
    conf.int = confidence_interval,
    stderr = sigma_hat / sqrt(n),
    alternative = alternative,
    weights = weights,
    data.name = d_name,
    method = paste0("Weighted multirater Cohen's kappa (w = ", w_name, "; ",
      method," CI)")
  )

  class(return_value) = "htest"

  return_value

}

#' @export
#' @describeIn cohen_kappa Fleiss' kappa
fleiss_kappa = function(
  x,
  alternative = c("two.sided", "greater", "less"),
  conf_level = 0.95,
  weights = "unit",
  kappa = 0,
  method = c("arcsine", "fisher", "basic"),
  skip = FALSE,
  count = NULL) {

  d_name = deparse1(substitute(x))
  w_name = deparse1(substitute(weights))
  method = match.arg(method)

  if(skip) {
    estimate = function(x) {
      pwf = pwf_hat(x, w = weights, count = count)
      pwa = pwa_hat(x, w = weights)
      (pwa - pwf) / (1 - pwf)
    }
    return(estimate(x))
  }

  alternative = match.arg(alternative)

  n = nrow(x)
  hats = var_kw(x, w = weights, type = "fleiss", count = count)
  sigma_hat = sqrt(hats$sigma_hat)
  kappa_hat = hats$kappa_hat

  # Make confidence interval
  if(alternative == "two.sided") {
    multiplier = c(stats::qt((1 - conf_level)/2, n - 1), -stats::qt((1 - conf_level)/2, n - 1))
  } else if (alternative == "greater") {
    multiplier = c(stats::qt((1 - conf_level), n - 1), Inf)
  } else if (alternative == "less") {
    multiplier = c(-Inf, -stats::qt((1 - conf_level), n - 1))
  }

  if(method == "basic") {
    confidence_interval = kappa_hat + sigma_hat / sqrt(n - 1) * multiplier
    confidence_interval[2] = min(1, confidence_interval[2])
  } else if (method == "arcsine") {
    std = sigma_hat / sqrt(1 - kappa_hat^2)
    confidence_interval = sin(asin(kappa_hat) + std / sqrt(n - 1) * multiplier)
  } else if (method == "fisher") {
    std = sigma_hat / (1 - kappa_hat^2)
    confidence_interval = tanh(atanh(kappa_hat) + std / sqrt(n - 1) * multiplier)
  }

  confidence_interval[2] = min(1, confidence_interval[2])
  attr(confidence_interval, "conf.level") = conf_level
  attr(confidence_interval, "alternative") = alternative
  attr(confidence_interval, "method") = method

  # Make statistic in p-value.
  statistic = sqrt(n) / sigma_hat * (kappa_hat - kappa)
  p_value = stats::pnorm(-abs(statistic)) / 2


  return_value = list(
    statistic = c(t = statistic),
    p.value = p_value,
    estimate = c("Fleiss's kappa" = kappa_hat),
    null.value = c(kappa = kappa),
    conf.int = confidence_interval,
    stderr = sigma_hat / sqrt(n),
    alternative = alternative,
    weights = weights,
    data.name = d_name,
    method = paste0("Weighted multirater Fleiss's kappa (w = ", w_name, "; ",
                    method," CI)")
  )

  class(return_value) = "htest"

  return_value

}
#' Estimate the sample weighted agreement coefficients
#'
#' The sample weighted agreement coefficients include ordinary weighted
#'    agreement along with two kinds of chance agreement: The Cohen-type and
#'    Fleiss-type. The Cohen-type of chance agreement does not count a judge's
#'    agreement with himself as chance agreement, while the Fleiss-type does.
#'
#' @keywords internal
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge.
#' @param w A weighting function. Can be choosen `"unit"`, `"quadratic"`,
#'    `"absolute_value"`.
#' @param count If `TRUE`, uses the counting algorithm that is faster for
#'    categorical data. Defaults to `NULL`, which evaluates to `TRUE` if the
#'    number of categories in the data is less than `sqrt(n)`, where `n` is the
#'    number of rows.
#' @return The sample weighted agreement, sample Cohen-type chance agreement,
#'   or sample Fleiss-type chance agreement.
pwa_hat = function(x, w = "unit") mean(mua_hats(x, w))

#' @describeIn pwa_hat The sample Cohen-type chance agreement.
pwc_hat = function(x, w = "unit", count = NULL) mean(muc_hats(x, w, count))

#' @describeIn pwa_hat The sample Fleiss-type chance agreement.
pwf_hat = function(x, w = "unit", count = NULL) mean(muf_hats(x, w, count))

#' The `mus` used in the calculation of the variance.
#'
#' Calculate `mus` for the sample weighted agreement, the sample Fleiss-type
#'   chance agreement, and the sample Cohen-type chance agreement. When `w` is
#'   a string, `C++` implementations ares used. When `w` is an `R` function,
#'   we can't use `C++`, and the evaluation is much slower.
#'
#' @keywords internal
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge.
#' @param w A weighting function. Can be choosen from the weights in `.weights`,
#'    or be a self-made function.
#' @return The `mus` used in the calculation of the variance.
#' @name muhats

#' @rdname muhats
mua_hats = function(x, w = "unit") {
  if(is_w_string(w)) {

    mua_hats_cpp_fun = function(x, w) {
      indices = arrangements::combinations(ncol(x), 2, replace = FALSE)
      index_u = indices[, 1] - 1
      index_v = indices[, 2] - 1
      c(mua_hats_cpp(x = x, w = w, index_u = index_u, index_v = index_v))
    }

    mua_hats_cpp_fun(x, w)

  } else {

    j = ncol(x)
    indices = arrangements::combinations(ncol(x), 2, replace = FALSE)
    n = nrow(x)
    X = seq(n)
    FUN = Vectorize(function(i) mean(w(x[i, indices[, 1]], x[i, indices[, 2]])))
    sapply(X, FUN)

  }
}

#' @rdname muhats
muc_hats = function(x, w = "unit", count = NULL) {

  indices = arrangements::combinations(ncol(x), 2, replace = FALSE)
  indices = rbind(indices, cbind(indices[, 2], indices[, 1]))

  if(is.null(count)) {

    if (length(unique(c(x))) < sqrt(nrow(x))) {
      count = TRUE
    } else {
      count = FALSE
    }

  }

  if(count) {
    if(is_w_string(w)) w = string_to_function(w)

    muc_hats_count = function(x, w) {

      cats = sort(unique(c(x)))
      w_matrix = outer(cats, cats, w)
      judges = ncol(x)
      n = nrow(x)

      # Subtracting 1 removes the junk added by cats. We need to add this to
      # make Table return a matrix, not a list.
      ncj = apply(x, 2, function(y) Rfast::Table(c(y, cats))) - 1
      ncj = rowSums(ncj) - ncj

      indices = cats
      vec = rep(0, n)
      for(j in seq(judges)) {
        vec = vec + sapply(match(x[, j], cats), function(k) w_matrix[k, ] %*% ncj[, j])
      }

      vec / (n * judges  * (judges  - 1))

    }

    return(muc_hats_count(x, w))

  }

  if(is_w_string(w)) {

    muc_hats_cpp_fun = function(x, w) {
      index_u = indices[, 1] - 1
      index_v = indices[, 2] - 1
      c(muc_hats_cpp(x = x, w = w, index_u = index_u, index_v = index_v))
    }

    muc_hats_cpp_fun(x, w)
  } else {
    j = ncol(x)
    n = nrow(x)
    X = Y = seq(n)
    FUN = Vectorize(function(i, j) mean(w(x[i, indices[, 1]], x[j, indices[, 2]])))
    c_matrix = outer(X = X, Y = Y, FUN = FUN)
    rowMeans(c_matrix)
  }
}

#' @rdname muhats
muf_hats = function(x, w = "unit", count = NULL) {

  indices = arrangements::combinations(ncol(x), 2)
  indices = rbind(indices, cbind(indices[, 2], indices[, 1]))
  indices = rbind(indices, cbind(seq(ncol(x)), seq(ncol(x))))

  if(is.null(count)) {

    if (length(unique(c(x))) < sqrt(nrow(x))) {
      count = TRUE
    } else {
      count = FALSE
    }

  }

  if(count) {
    if(is_w_string(w)) w = string_to_function(w)

    muf_hats_count = function(x, w) {
      cats = sort(unique(c(x)))
      x_ = match(x, cats)
      dim(x_) = dim(x)
      counts = Rfast::Table(x = rep(seq(nrow(x_)), ncol(x_)),y = c(x_))
      w_matrix = outer(cats, cats, w)
      j = ncol(x)
      n = nrow(x)
      ncs = colSums(counts)
      unname(c(apply(counts, 1, function(row) row %*% w_matrix %*% ncs) / (n * j^2)))
    }

    return(muf_hats_count(x, w))

  }

  if(is_w_string(w)) {

    muf_hats_cpp_fun = function(x, w) {
      index_u = indices[, 1] - 1
      index_v = indices[, 2] - 1
      c(muc_hats_cpp(x = x, w = w, index_u = index_u, index_v = index_v))
    }

    return(muf_hats_cpp_fun(x, w))

  } else {
    j = ncol(x)
    n = nrow(x)
    X = Y = seq(n)
    FUN = Vectorize(function(i, j) mean(w(x[i, indices[, 1]], x[j, indices[, 2]])))
    f_matrix = outer(X = X, Y = Y, FUN = FUN)
    rowMeans(f_matrix)

  }


}

#' Variances of Cohen's kappa and Fleiss' kappa
#'
#' @keywords internal
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge.
#' @param w A weighting function. Can be choosen `"unit"`, `"quadratic"`,
#'    `"absolute_value"`.
#' @return Variances of Cohen's kappa and Fleiss' kappa
var_kw = function(x, w = "unit", type = c("cohen", "fleiss"), count = NULL) {
  type = match.arg(type)

  muas = mua_hats(x, w)
  mucs = if (type == "cohen") muc_hats(x, w, count) else muf_hats(x, w, count)
  pwa = mean(muas)
  pwc = mean(mucs)
  sigmas = c(stats::var(muas), 4 * stats::cov(muas, mucs), 4 * stats::var(mucs))

  qwc = 1/(1-pwc)
  qwa = pwa - 1
  ps = c(1, qwa * qwc, qwa ^2 * qwc ^2) * qwc^2

  list(kappa_hat = (pwa - pwc) / (1 - pwc), sigma_hat = c(ps %*% sigmas))

}
