#' Estimate a knowledge coefficient.
#'
#' The estimands of the coefficients in this function are approximations to the
#'    knowledge coefficient, defined for guessing models. The approximations are
#'    exact in some circumstances. They are similar to chance-corrected
#'    measures of agreement, and are defined outside the context of guessing
#'    models too.
#'
#' Most of the coefficient admits both weighted and unweighted variants, the
#'    exceptions being `bp` and `meancorr`.
#'
#' Standard deviations and confidence sets are calculated using linearized
#'    variance and covariance estimates. They are valid also outside the context
#'    of guessing models.
#'
#' Model `bp` returns the Brennan-Prediger coefficient, and is not affected by
#'    weights. It's equivalent to `irrCAC::bp.raw`.
#'
#' Model `fleiss` returns the weighted Fleiss' kappa. It's equivalent to
#'    `irrCAC::fleiss.kappa.raw`.
#'
#' Model `cohen` returns the weighted multirater Cohen kappa.
#'
#' Model `cf` returns the weighted Cohen-Fleiss kappa.
#'
#' Model `cbp` returns the weighted Cohen-Brennan-Prediger kappa.
#'
#' @param x An array of raw data. Each row corresponds to an item and
#'    each column corresponds to a judge.
#' @param alternative a character string specifying the alternative hypothesis,
#'    must be one of `"two.sided"` (default), `"greater"` or `"less"`. Used both
#'    for the test and the confidence interval.
#' @param conf_level Confidence level for the confidence interval. Defaults to
#'    to `0.95`.
#' @param model The model used in the estimation of the coefficient. The
#'    `"bp"`model assumes that the guessing distribution is uniform, the
#'    `"fleiss"` model assumes that the guessing distribution equals the
#'    true distribution of ratings.
#' @param weights Choice of weights.
#' @param nu The skill coefficient of the null-hypothesis. Defaults to `0`,
#'    which is the minimal value the skill coefficient.
#' @param skip If `TRUE`, skips inference and calculates the skill
#'    coefficient only. Defaults to `FALSE`
#' @param cat Number of categories. Defaults the number of unique values in the
#'    data.
#' @return The skill coefficient coefficient.
#' @export
skill = function(
  x,
  alternative = c("two.sided", "greater", "less"),
  conf_level = 0.95,
  model = c("bp", "fleiss", "cf", "cohen", "cbp"),
  weights = "unit",
  nu = 0,
  skip = FALSE,
  cat = length(unique(c(x)))) {

  d_name = deparse1(substitute(x))
  model = match.arg(model)

  estimate = function(x) {

    if(model == "bp") {

      pa = pwa_hat(x, w = "unit")
      top = bottom = 1/cat

    } else {

      pa = pwa_hat(x, w = weights)

      if(model == "fleiss") {
        top = bottom = pwf_hat(x, w = weights)
      } else if(model == "bp") {
        pa = pwa_hat(x, w = weights)
        top = bottom = 1/cat
      } else if(model == "cbp") {
        top = pwc_hat(x, w = weights)
        bottom = 1 / cat
      } else if(model == "cf") {
        top = pwc_hat(x, w = weights)
        bottom = pwf_hat(x, w = weights)
      } else if(model == "cohen") {
        top = bottom = pwc_hat(x, w = weights)
      }

    }

    (pa - top) / (1 - bottom)

  }

  if(skip) return(estimate(x))

  if(model == "bp") {
    alternative = match.arg(alternative)
    pi_hat = table(x) / length(x)
    n = nrow(x)

    pa = pwa_hat(x, w = "unit")

    # Estimate.
    nu_hat = estimate(x)

    tau = stats::sd(mua_hats(x))  *  cat / (cat - 1)

    # Make confidence interval
    multiplier = c(stats::qt((1 - conf_level)/2, n - 1), -stats::qt((1 - conf_level)/2, n - 1))
    confidence_interval = nu_hat + tau / sqrt(n - 1) * multiplier
    attr(confidence_interval, "conf.level") = conf_level

    # Make statistic in p-value.
    statistic = sqrt(n) / tau * (nu_hat - nu)
    p_value = stats::pnorm(-abs(statistic)) / 2

    return_value = list(
      statistic = c(t = statistic),
      p.value = p_value,
      estimate = c("Brennan-Prediger coefficient (nu)" = nu_hat),
      null.value = c(nu = nu),
      conf.int = confidence_interval,
      stderr = tau / sqrt(n),
      alternative = alternative,
      method = paste0("Brennan-Prediger coefficient."),
      data.name = d_name
    )

    class(return_value) = "htest"

    return_value
  }

}


