#' Simulate from a multiplicative guessing model.
#'
#' @param n The number of items to sample.
#' @param skills Vector of skill parameters. Recycled to match the number of
#'     judges. If `1`, the the model is an item difficulty model.
#' @param difficulties The item difficulties. A function of `n` that generates
#'    difficulty parameters from `0` to `1`, e.g. `runif`. Parameters may be
#'    passed to difficulties using `...` If `NULL`, the difficulties are
#'    constant and the model is a judge skill model.
#' @param true_dist A vector of true category probabilities.
#' @param guessing_dist An array where each row contains the guessing
#'    distribution for the jth judge, or a vector containing a single
#'    guessing distribution. Recycled to match the number of judges.
#' @param judges The (optional) number of judges. Defaults to the maximum of
#'    `length(skills)` and the rows of `guessing_dist`. Recycles guessing `skills`
#'    and `guessing_dist`.
#' @param ... Passed to `difficulties`.
#' @return Simulated `raw` data, a matrix with `judges` columns and `n` rows.
#' @export

simulate_mgm = function(
  n,
  skills = 1,
  difficulties = NULL,
  true_dist,
  guessing_dist,
  judges,
  ...) {

  if(missing(judges)) judges = max(length(skills), length(guessing_dist) / length(true_dist))
  skills = rep_len(skills, judges)
  q = length(true_dist)

  stopifnot(judges > 1)
  if(length(guessing_dist) != judges * q) {
    guessing_dist = matrix(rep_len(guessing_dist, judges * q),
                           nrow = judges, byrow = TRUE)
  }

  q = length(true_dist)
  x_star = sample(q, n, replace = TRUE, prob = true_dist)

  observations = sapply(seq_along(skills), function(j) {
    z = stats::rbinom(n, 1, skills[j])
    z * x_star + (1 - z) * sample(q, n, replace = TRUE, guessing_dist[j, ])
  })

  if(is.null(difficulties)) {
    d_samples = matrix(rep(skills, n), ncol = judges, byrow = TRUE)
  } else {
    d_samples = difficulties(n, ...) * matrix(rep(skills, n), ncol = judges, byrow = TRUE)
  }

  z = matrix(stats::rbinom(n * judges, 1, c(d_samples)), ncol = judges)

  samples = sapply(seq(nrow(guessing_dist)), function(j) {
    sample(q, n, replace = TRUE, guessing_dist[j, ])
  })

  observations = z * x_star + (1 - z) * samples

  attr(observations, "n") = n
  attr(observations, "skills") = skills
  attr(observations, "difficulties") = difficulties
  attr(observations, "true_dist") = true_dist
  attr(observations, "guessing_dist") = guessing_dist
  observations

}

#' Simulate from a judge skill model.
#'
#' @param n The number of observations to sample.
#' @param s Vector of success probabilities.
#' @param model Which kind of guessing model to simulate from. This determines
#'    the shape of the guessing distribution. Choose from `"general`, which uses
#'    both true_dist and guessing_dist directly,  `"bp"`
#'    (uniform guessing distribution)`"fleiss"` (guessing distribution
#'    equal to the true distribution), `"cohen` (marginal distribution equal to true
#'    distribution), and `"tu"` (true distribution equal to the uniform distribution.)
#' @param true_dist A vector of true category probabilities. Ignored when `model = "cohen"`
#'    or `model = "tu`.
#' @param guessing_dist An array where each row contains the guessing
#'    distribution for the jth judge, or a vector containing a single
#'    guessing distribution.  Ignored when `model = "bp"` or `model = "fleiss"`.
#' @return Simulated `raw` data.
#' @export

simulate_jsm = function(
  n,
  s,
  model = c("general", "cohen-fleiss", "fleiss", "bp", "tu"),
  true_dist,
  guessing_dist) {

  model = match.arg(model)

  if(model == "cohen-fleiss") {
    if(is.null(nrow(guessing_dist))) {
      true_dist = guessing_dist
    } else {
      true_dist = colSums(guessing_dist * (1 - s) / sum(1 - s))
    }
  } else if(model == "tu") {
    true_dist = rep(1, ncol(guessing_dist)) / ncol(guessing_dist)
  } else if (model == "bp") {
    guessing_dist = rep(1, length(true_dist)) / length(true_dist)
  } else if (model == "fleiss") {
    if(any(true_dist < 0)) stop("true_dist must contain only positive elements.")
    guessing_dist  = true_dist / sum(abs(true_dist))
  }

  # The true category for each item.
  q = length(true_dist)
  x_star = sample(q, n, replace = TRUE, prob = true_dist)

  if(is.null(nrow(guessing_dist))) {
    observations = sapply(seq_along(s), function(j) {
      z = stats::rbinom(n, 1, s[j])
      z * x_star + (1 - z) * sample(q, n, replace = TRUE, guessing_dist)
    })
  } else {
    observations = sapply(seq_along(s), function(j) {
      z = stats::rbinom(n, 1, s[j])
      z * x_star + (1 - z) * sample(q, n, replace = TRUE, guessing_dist[j, ])
    })
  }

  attr(observations, "n") = n
  attr(observations, "s") = s
  attr(observations, "true_dist") = true_dist
  attr(observations, "guessing_dist") = guessing_dist
  attr(observations, "skill") = true_skill(s)
  observations

}


#' Simulate from an item difficulty model
#'
#' In the generalized Aickin model, the skill-difficulties s are constand across
#'    judges but may vary across items. The judges may have different guessing
#'    distributions.
#'
#' In Aickin's original model, the true distribution is equal to the normalized
#'    geometric mean of the guessing distributions. Moreover, the skill-difficulty
#'    parameters `s(1)` are binary, i.e. `s = 0` or `s(1) = 1`.
#'
#' @param n Number of observations to sample.
#' @param d A function of `n` that generates difficulty parameters from `0` to `1`,
#'    e.g. `runif`.
#' @param model If `"aickin"`, defines `true_dist` to be equal to the
#'    normalized geometric mean of the guessing distributions.
#' @param true_dist A vector of true category probabilities. Ignored when `model = "aickin"`.
#' @param guessing_dist An array of guessing distributions, one row for every
#'    judge.
#' @param ... Passed to `d`.
#' @return Simulated `raw` data.
#' @export
#' @examples
#'  guessing_dist = rbind(
#'    c(0.05, 0.05, 0.5, 0.4),
#'    c(0.3, 0.3, 0.1, 0.3),
#'    c(0.3, 0.3, 0.1, 0.3)
#'    )
#'
#'  n = 30
#'
#'  # Simulate from a classical Aickin model.
#'  x = simulate_idm(
#'       n,
#'       d = rbinom,
#'       model = "aickin",
#'       guessing_dist = guessing_dist,
#'       size = 1,
#'       prob = 0.3)
#'
#'  # Simulate from an Aickin model with specified true distribution and
#'  # skill parameters different from 0 and 1.
#'  n = 30
#'  true_dist = c(0.3, 0.3, 0.1, 0.3)
#'  y = simulate_idm(
#'       n,
#'       d = rbeta,
#'       model = "general",
#'       true_dist,
#'       guessing_dist,
#'       shape1 = 9,
#'       shape2 = 1)
#'
#' # The true skill-difficulty parameter is E(s^2) = 0.818181.

simulate_idm = function(n, d, model = c("general", "aickin"), true_dist, guessing_dist, ...) {


  model = match.arg(model)
  j = nrow(guessing_dist)

  if(model == "aickin") {
    true_dist = apply(guessing_dist, 2, prod)
    true_dist = true_dist / sum(true_dist)
  }

  q = length(true_dist)
  x_star = sample(q, n, replace = TRUE, prob = true_dist)
  s_ = d(n, ...)

  z = matrix(stats::rbinom(n * j, 1, s_), ncol = j)

  samples = sapply(seq(nrow(guessing_dist)), function(j) {
    sample(q, n, replace = TRUE, guessing_dist[j, ])
  })

  observations = z * x_star + (1 - z) * samples

  attr(observations, "n") = n
  attr(observations, "d") = d
  attr(observations, "true_dist") = true_dist
  attr(observations, "guessing_dist") = guessing_dist
  observations

}
