#' Unique values except NA.
#'
#' @param x A value that supports "c(x)".
#' @return The vector c(x) with NAs removed.

unique_remove_na = function(x) {
  x = unique(c(x))
  x[!is.na(x)]
}

#' Transform agreement data formats
#'
#' Agreement data comes in many formats. These utility functions transform between
#'   them, when possible.
#'
#'
#' * Cohen form (for two judges; a contingency table) (irrCAC::cont3x3abstractors, irrCAC::cont4x4diagnosis)
#' * Krippendorff form (`table` of pair form; discard information about items and judges).
#' * Fleiss form (item rows; number of ratings for each category) Discards information about the judges. (irrCAC::cac.dist4cat)
#' * Raw form / wide form (item rows; judges on columns) (cac.raw4raters)
#' * Triplet form / long form (x, j, i) This is usually the best form, as it discards no data, easily handles covariates, and multiple ratings.
#'
#' @param x The input data.
#' @param input Form of the input data. See the details.
#' @param output Form of the output data. See the details.
#' @param rating Optional name of rating column when transforming input data on
#'     triple form. Defaults to `"rating"`.
#' @return Agreement data on the `"output"` form, provided a transformation is
#'     possible.
transform_data = function(x, input, output, rating) NULL

#' Transforms triplet to paired data. (Remove?)
#' @param triplet Data on triplet form.
#' @return Data on paired form.

triplet_to_pair = function(triplet) {

  data_list = suppressWarnings(dplyr::group_split(dplyr::group_by_(triplet, "item")))
  data = cbind(rep(NA, nrow(triplet)^2), rep(NA, nrow(triplet)^2))
  index = 0

  for (mat in data_list) {
    if (nrow(mat) > 1) {
      indices = arrangements::combinations(seq(nrow(mat)), 2)
      for (i in seq(nrow(indices))) {
        data[index + i, ] = c(as.numeric(mat[indices[i, 1], 1]),
                              as.numeric(mat[indices[i, 2], 1]))
      }

      index = index + i

    }
  }

  data[(index + 1):(2 * index), ] = data[1:index, c(2, 1)]

  data[1:(2 * index), ]

}


#' Transform agreement data on raw data form to triplet form.
#'
#' Transforms data on the raw form to data on the normal triplet form.
#'
#' Data on raw form follows the convention of Gwet.
#'
#' @param x Agreement data in raw data frame form.
#' @return Agreement data on normal triplet form. This is a tibble
#' @export

raw_to_triplet = function(x) {
  size = sum(!is.na(x))
  triplet = matrix(NA, nrow = size, ncol = 3)
  colnames(triplet) = c("rating", "item", "judge")
  counter = 1

  for (j in seq(nrow(x))) {
    for (i in seq(ncol(x))) {
      if (!is.na(x[j, i])) {
        triplet[counter, 1] = x[j, i]
        triplet[counter, 2:3] = c(i, j)
        counter = counter + 1
      }
    }
  }

  tibble::as_tibble(triplet)

}

#' Transform agreement data on raw form to Krippendorff form.
#'
#' Data on raw form follows the convention of Gwet.
#'
#' @param x Agreement data on raw
#' @return Agreement data on raw as a Krippendorff matrix.
#' @export
raw_to_krippendorff = function(x) {
  krippendorff_matrix = matrix(0, nrow = cat, ncol = cat)
  for(j1 in seq(ncol(x) - 1)) {
    for(j2 in j1:ncol(x)) {
      for(i in seq(nrow(x))) {
        krippendorff_matrix[x[i, j1], x[i, j2]] = krippendorff_matrix[x[i, j1], x[i, j2]] + 1
        krippendorff_matrix[x[i, j2], x[i, j1]] = krippendorff_matrix[x[i, j2], x[i, j1]] + 1
      }
    }
  }

  krippendorff_matrix
}

# To transform to Fleiss form

# agree.mat <- matrix(0, nrow=n, ncol=q)
# for(k in 1:q){
#   categ.is.k <- (ratings.mat==categ[k])
#   agree.mat[,k] <- (replace(categ.is.k,is.na(categ.is.k),FALSE)) %*% rep(1,r)
# }
# agree.mat.w <- t(weights.mat%*%t(agree.mat))


#' Transform agreement data on triplet form to raw form.
#'
#' Data on raw form follows the convention of Gwet.
#'
#' @param x Agreement data on triplet form.
#' @param rating Optional name of rating column; defaults to `"rating"`.
#' @return Agreement data on raw as a matrix.
#' @export


triplet_to_raw = function(x, rating = "rating") {
  as.matrix(dplyr::select(tidyr::spread_(x, key = "judge", value = rating), -1))
}

#' Data formats
#'
#' Cohen form (for two judges)
#' Krippendorff form (`table` of pair form; discard information about items and judges).
#' Fleiss form (item rows; number of ratings for each category) Discards information about the judges.
#' Raw form / wide form (item rows; judges on columns)
#' Triplet form / long form (x, j, i) This is usually the best form, as it discards no data, easily handles covariates, and multiple ratings.

#cohen_to_triple = fun