#' Calculates theoretical coefficient.
#'
#' @param s Vector of success probabilities.
#' @param true_dist A vector of true category probabilities.
#' @param guessing_dist An array where each row contains the guessing
#'    distribution for the jth judge, or a vector containing a single
#'    guessing distribution.
#' @param type The model used in the estimation of the coefficient. The `"bp"`
#'    model assumes that the guessing distribution is uniform, the `"fleiss"`
#'    model assumes that the guessing distribution equals the marginal
#'    distribution of ratings.
#' @param ac1 If `TRUE`, calculates the AC1 under the model specified in type.
#'    Ignored if `type = "skill"`, `"pcac"`, `"pcaf"`, or `"pa"`.
#' @param approximation If `TRUE`, returns the approximation to the AC1 under the
#'     `"bp"` model. Ignored if `AC1 = FALSE`.
#' @param loss A bivariate function assigning an disagreement weight l(i, j) to
#'    every combination of categories. Currently not supported.
#' @return The true coefficient.
#' @export
true_coefficient = function(
  s,
  type = c("skill", "cohen-fleiss", "fleiss", "bp", "tu", "pcac", "pcaf", "pa"),
  true_dist,
  guessing_dist,
  ac1 = FALSE,
  approximation = FALSE,
  loss) {

  type = match.arg(type)

  if (type == "skill") {
    return(true_skill(s))
  } else if (type == "pcac") {
    return(true_pcac(s, true_dist, guessing_dist))
  } else if (type == "pcaf") {
    return(true_pcaf(s, true_dist, guessing_dist))
  } else if (type == "pa") {
    return(true_pa(s, true_dist, guessing_dist))
  }

  if(!ac1) {
    if(type == "fleiss") {
      return(true_fleiss(s, true_dist, guessing_dist))
    } else if(type == "bp") {
      return(true_bp(s, true_dist, guessing_dist))
    } else if(type == "tu") {
      #return(true_tu(s, true_dist, guessing_dist))
    } else if(type == "cohen-fleiss") {
      #return(true_cohen_fleiss(s, true_dist, guessing_dist))
    }
  }

}

true_skill = function(s) {
  (sum(outer(s, s)) - sum(s * s)) / 2 / choose(length(s), 2)
}

true_pa = function(s, true_dist, guessing_dist) {
  ss = apply(utils::combn(length(s), 2), 2, function(i) s[i[1]] * s[i[2]])
  qq = c(guessing_dist %*% guessing_dist)
  tq = c(true_dist %*% guessing_dist)
  pa = mean(ss * (1 + qq - 2 * tq)) + 2 * (tq - qq) * mean(s) + qq
}

true_pcaf = function(s, true_dist, guessing_dist) {
  ss = apply(utils::combn(length(s), 2), 2, function(i) s[i[1]] * s[i[2]])
  qq = c(guessing_dist %*% guessing_dist)
  tq = c(true_dist %*% guessing_dist)
  tt = c(true_dist %*% true_dist)
  mean(ss) * (tt - 2 * tq + qq) + 2 * mean(s) * (tq - qq) + qq
}

true_pcac = function(s, true_dist, guessing_dist) {
  qq = c(guessing_dist %*% guessing_dist)
  tq = c(true_dist %*% guessing_dist)
  tt = c(true_dist %*% true_dist)
  true_skill(s) * (tt - 2 * tq + qq) + 2 * mean(s) * (tq - qq) + qq
}

true_ac1 = function(s, true_dist, approximation = FALSE, model = c("bp", "fleiss")) {

  model = match.arg(model)
  q = length(true_dist)

  if(approximation) {
    p_a = true_skill(s) * (1 - 1/q) + 1/q
    p_e = 1/q - 1/q * mean(s)^2 * (q * sum(true_dist^2) - 1) / (q - 1)
    (p_a - p_e) / (1 - p_e)

  } else {

    if(model == "bp") {
      p_a = true_skill(s) * (1 - 1/q) + 1/q
    } else {
      p_ca = sum(true_dist^2)
      p_a = true_skill(s) * (1 - p_ca) + p_ca
    }

    p_e = p_a - true_skill(s)
    (p_a - p_e) / (1 - p_e)

  }

  (p_a - p_e) / (1 - p_e)

}

true_bp = function(s, true_dist, guessing_dist) {
  pa = true_pa(s, true_dist, guessing_dist)
  pga = 1 / length(true_dist)
  (pa - pga) / (1 - pga)
}

true_fleiss = function(s, true_dist, guessing_dist) {
  probs = mean(s) * true_dist + (1 - mean(s)) * guessing_dist
  pga = c(probs %*% probs)
  pa = true_pa(s, true_dist, guessing_dist)
  (pa - pga) / (1 - pga)
}
