
#' Calculates confidence interval
#'
#' @param est,sd The estimate and estimated standard deviation.
#' @param n Number of observations.
#' @param method The method used in the confidence interval construction.
#' @param conf_level Confidence level.
#' @param alternative Direction of the confidence interval.
#' @keywords internal

ci <- function(est, sd, n, method, conf_level, alternative) {
  multiplier <- if (alternative == "two.sided") {
    c(
      stats::qt((1 - conf_level) / 2, n - 1),
      -stats::qt((1 - conf_level) / 2, n - 1)
    )
  } else if (alternative == "greater") {
    c(stats::qt((1 - conf_level), n - 1), Inf)
  } else if (alternative == "less") {
    c(-Inf, -stats::qt((1 - conf_level), n - 1))
  }
  if (method == "basic") {
    std <- sd
    ci <- est + sd / sqrt(n - 1) * multiplier
    ci[2] <- min(1, ci[2])
  } else if (method == "arcsine") {
    std <- sd / sqrt(1 - est^2)
    tmp <- asin(est) + std / sqrt(n - 1) * multiplier
    ci <- c(0, 0)
    if (tmp[1] == -Inf) {
      ci[1] <- -1
    } else {
      ci[1] <- sin(tmp[1])
    }
    if (tmp[2] == Inf) {
      ci[2] <- 1
    } else {
      ci[2] <- sin(tmp[2])
    }
  } else if (method == "fisher") {
    std <- sd / (1 - est^2)
    ci <- tanh(atanh(est) + std / sqrt(n - 1) * multiplier)
  }
  ci[2] <- min(1, ci[2])
  attr(ci, "conf.level") <- conf_level
  attr(ci, "alternative") <- alternative
  attr(ci, "method") <- method
  ci
}

#' Calculates p-value
#'
#' @param statitics The statistic used.
#' @param n Number of observations.
#' @param alternative Alternative hypothesis.
#' @keywords internal
p_value <- function(statistic, n, alternative) {
  if (alternative == "two.sided") {
    stats::pt(-abs(statistic), n - 1) / 2
  } else if (alternative == "greater") {
    stats::pt(statistic, n - 1)
  } else if (alternative == "less") {
    stats::pt(-statistic, n - 1)
  }
}

#' Prepare data
#'
#' Change data from wide form to Fleiss form if need be. Check for `NA`s. Stop
#' computation if Cohen's kappa is attempted on Fleiss form data.
#'
#' @param x The statistic used.
#' @param fleiss_form Number of observations.
#' @param type Alternative hypothesis.
#' @keywords internal
prepare_data <- function(x, fleiss_form, type) {

  if (any(is.na(x))) {
    stop("NA data is not allowed in rectangular designs.")
  }

  if (fleiss_form) {
    if (type == "cohen") {
      stop("Impossible to calculate Cohen's kappa with data on Fleiss form.")
    }
    x <- disaggr(x)
  } else {
    x <- as.matrix(x)
  }

  x

}

#' Prepare g
#'
#' Handele `NULL` gs. If `disagreement` is `hubert`, modify it to the number of
#'    raters. If not, modify it to `2`. Only works with data on wide form.
#'
#' @param g The supplied value of `g`.
#' @param disagreement Disagreement string.
#' @param ncol Number of raters
#' @return The correct value of g.
#' @keywords internal
prepare_g <- function(g, disagreement, ncol) {
  if (is.null(g) & disagreement == "hubert") {
    g <- ncol
  }
  ifelse(is.null(g), 2, g)
}

#' Construct an htest function from data.
#' @keywords internal
construct_htest <- function(kappa_hat, type, g, disagreement, method, statistic,
                            p_value, sigma_hat, kappa, ci, n, alternative, d_name,
                            fleiss_form, ncol) {
  descriptions <- list("fleiss" = "Fleiss' kappa", "cohen" = "Cohen's kappa")
  names(kappa_hat) <- c(descriptions[type])

  method_desc <- if (g == ncol & disagreement == "hubert") {
    paste0("Hubert's kappa (", method, " CI)")
  } else {
    paste0(
      g, "-wise ", descriptions[type],
      " ('", disagreement, "' disagreement; ", method, " CI)"
    )
  }

  return_value <- list(
    statistic = c(t = statistic),
    p.value = p_value,
    estimate = c("kappa" = unname(kappa_hat), "std" = sigma_hat),
    null.value = c(kappa = kappa),
    conf.int = ci,
    stderr = sigma_hat / sqrt(n),
    alternative = alternative,
    disagreement = disagreement,
    data.name = paste0(d_name, " (items: ", n, "; raters: ", ncol, "); "),
    method = method_desc
  )

  class(return_value) <- "htest"

  return_value
}

#' Transform data Fleiss form to fake data on wide form.
#'
#' Data on Fleiss form doesn't keep which judge made which rating. Instead, it
#'    aggregates them into counts. This function disaggregates the data into
#'    one of the possible sources. There are infinitely many, but that doesn't
#'    matter for the computation of Fleiss' kappa.
#'
#' @param x Data on the Fleiss form.
#' @keywords internal
#' @return Fake wide form data.
disaggr <- function(x) {
  x <- as.matrix(x)
  y <- matrix(nrow = nrow(x), ncol = sum(x[1, ]))
  for (row in seq(nrow(y))) {
    current <- 1
    for (col in seq(ncol(y))) {
      while (TRUE) {
        if (x[row, current] != 0) {
          y[row, col] <- current
          x[row, current] <- x[row, current] - 1
          break
        } else {
          current <- current + 1
        }
      }
    }
  }
  y
}

#' Aggregate wide form data.
#' @keywords internal
aggr <- \(x, cats = NULL) {
  if(is.null(cats)) cats <- sort(unique(c(as.matrix(x))))
  x <- as.matrix(x)
  results <- matrix(0, nrow(x), length(cats))
  for (i in seq_len(nrow(x))) {
    counts = Rfast::Table(as.matrix(x)[i, ])
    results[i, as.integer(names(counts))] = counts
  }
  results
}
