
<!-- README.md is generated from README.Rmd. Please edit that file -->

# agreer <img src="man/figures/logo.png" align="right" width="217" height="142" />

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/agreer)](https://cran.r-project.org/package=agreer)
[![Project Status: WIP – Initial development is in progress, but there
has not yet been a stable, usable release suitable for the
public.](https://www.repostatus.org/badges/latest/wip.svg)](https://www.repostatus.org/#wip)
[![R-CMD-check](https://github.com/JonasMoss/agreer/workflows/R-CMD-check/badge.svg)](https://github.com/JonasMoss/agreer/actions)
[![Codecov test
coverage](https://codecov.io/gh/JonasMoss/agreer/branch/main/graph/badge.svg)](https://app.codecov.io/gh/JonasMoss/agreer?branch=main)
<!-- badges: end -->

This is an `R` package for estimating and doing inference with measures
of agreement, also known as measures of inter-rater agreement,
implementing the methods of Moss (2022a) and Moss (2022b). It supports
asymptotically valid inference, such as confidence intervals and
*p*-values, for a wide variety of agreement coefficients in a unified
interface.

1.  **Fixed or exchangeable raters.** The chance-corrected agreement
    coefficients assume that all ratings are conditionally independent
    given the item, but Cohen’s kappa-like statistics assume *fixed
    raters*, and Fleiss’ kappa-like statistics assume *exchangeable
    raters* (conditioned on the item). It is possible to handle missing
    data with exchangeable raters under minimal assumptions.
2.  ***g*-wise agreement coefficients for any *g* less than or equal to
    R**. Most known agreement coefficients are defined in a pairwise
    manner, with Hubert’s kappa being the best known exception. But
    there are good reasons to use another *g* if that is your use case.
    When the raters are fixed you probably want to compute the agreement
    among exactly these raters. The function `agree` automatically
    chooses the number of raters `R` when the option `raters = "fixed"`
    is chosen. But *g*s different from `R` and `2` are useful too. For
    example, if you use the median judgement of 5 raters in a regression
    model, the most suitable agreement coefficient is the the 5-wise
    agreement coefficient with fixed raters and the `absolute`
    disagreement function. For this is the agreement coefficient that
    measures how much they agree on their aggregate judgement.
3.  **Four classes disagreement functions.** The Hubert disagreement
    function counts a vector of ratings as agreeing if and only if all
    raters agree. The arguments `"nominal"`, `"absolute"` and
    `"quadratic"` are three *Fréchet variances* generalizing the
    nominal, absolute and quadratic losses to *g \> 2*. See the vignette
    for a detailed explanation of these disagreement functions.

When the raters are fixed, the agreement coefficients are defined only
for data where every item is rated by the same number of judges, and if
a judge fails to rate an item it must be thrown out. If the judges are
exchangeable given the item rated it is possible to do valid inference
without throwing away any data (Moss, 2022b), provided the data is
missing completely at random.

Most well-known agreement coefficients fit into the framework.

| Name                   | Raters | *g* | Disagreement | Raters       | Missing data? |
|------------------------|--------|-----|--------------|--------------|---------------|
| Cohen’s kappa          | 2      | 2   | Nominal      | Fixed        | No            |
| Weighted Cohen’s kappa | 2      | 2   | Any          | Fixed        | No            |
| Conger’s kappa         | Any    | 2   | Nominal      | Fixed        | No            |
| Schuster-Smith coef    | Any    | Any | Any          | Fixed        | No            |
| Scott’s pi             | 2      | 2   | Nominal      | Exchangeable | Yes!          |
| Fleiss’ kappa          | Any    | 2   | Nominal      | Exchangeable | Yes!          |
| Weighted Fleiss’ kappa | Any    | 2   | Any          | Exchangeable | Yes!          |
| Krippendorff’s alpha\* | Any    | 2   | Any          | Exchangeable | Yes!          |
| Hubert’s kappa\*\*     | Any    | R   | Hubert       | Exchangeable | No            |

## Installation

Install the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("remotes")
remotes::install_github("JonasMoss/agreer")
```

## Example usage

Zapf et al. (2016) (`dat.zapf2016`) did an agreement case study of
histopathological assessment of breast cancer. It involved ratings done
by R=4 senior pathologists and n=50 breast cancer biopsies.

``` r
library("agreer")
#> 
#> Attaching package: 'agreer'
#> The following object is masked from 'package:base':
#> 
#>     kappa
head(dat.zapf2016)
#>   Rater A Rater B Rater C Rater D
#> 1       5       5       4       5
#> 2       1       1       1       1
#> 3       5       5       5       5
#> 4       1       3       3       3
#> 5       5       5       5       5
#> 6       1       1       1       1
```

The rows are the items rated and the columns are the raters. The raters
are fixed in this study, so we use the argument Then we calculate the
4-way agreement coefficient for fixed raters using the absolute value
weights and the recommended arcsine confidence interval.

``` r
kappa(dat.zapf2016, disagreement = "absolute", g = 4)
#> 
#>  4-wise Cohen's kappa ('absolute' disagreement; arcsine CI)
#> 
#> data:  dat.zapf2016 (items: 50; raters: 4); 
#> t = 20.494, p-value < 2.2e-16
#> alternative hypothesis: true kappa is not equal to 0
#> 95 percent confidence interval:
#>  0.7126902 0.8704058
#> sample estimates:
#>     kappa       std 
#> 0.7984257 0.2754787
```

We can also calculate agreement coefficients for exchangeable raters by
using the `raters = "exchangeable"` argument. We use `g = 6`, which
equals all the raters in the data set. For an explanation about what
happens for the `"nominal"` disagreement with multiple raters, see the
vignette.

``` r
kappa(dat.fleiss1971, type = "fleiss", disagreement = "nominal", g = 6, fleiss_form = TRUE)
#> 
#>  6-wise Fleiss' kappa ('nominal' disagreement; arcsine CI)
#> 
#> data:  dat.fleiss1971 (items: 30; raters: 6); 
#> t = 8.7355, p-value = 3.226e-10
#> alternative hypothesis: true kappa is not equal to 0
#> 95 percent confidence interval:
#>  0.3663360 0.5971122
#> sample estimates:
#>     kappa       std 
#> 0.4859784 0.3047108
```

## Notes

(\*) Krippendorff’s alpha is a small modification of the suboptimal
exhangeable agreement coefficient with equal weighthing function. You
can use it by selection `krippendorff_modification = TRUE`, and
`w = "linear"`, but there is little reason to do so, as the resulting
estimator is worse than the default `w = "first-order"`.

(\*\*) `agreer` implements the Hubert disagreement for any *g*. When *g*
isn’t equal to `R`, the number of raters, it takes the mean of the
*g*-wise Hubert disagreement instead.

## How to help and todos

This project has a code of conduct here. There are several things to do.

- Implement functions that allows for vector-valued agreement data,
  probably on long form.
- Implement a transformation function that convert between wide form and
  long form, and from wide form to Fleiss form.
- Write a vignette about the different data forms and which forms are
  suitable for what.
- Clean up the `C++` code; port more of the functionality to `C++`. That
  would make it easier to port whole library to Python, SPSS, STATA, web
  apps, and so on.
- Write code that works better with continuous data. The current code is
  fast for data with few categories, but is perhaps too slow with
  continuous data. This could plausibly be ameliorated by doing Monte
  Carlo sampling.
- Write a vignette explanining the connection between Krippendorff’s
  alpha and the exchangeable agreement coefficient when *g* is 2.

Please contact us if you’re interested in contributing, and don’t
hesitate to open pull requests.

## References

- Moss, Jonas. “Weighted chance-corrected measures of agreement with
  multiple raters.” Submitted (2022a)
- Moss, Jonas. “Inference for g-wise Fleiss’ kappa and Krippendorff’s
  alpha-like statistics with missing data” Submitted (2022b)
- Zapf, Antonia, Castell, Stefanie, Morawietz, Lars, and Karch, André,
  “Measuring inter-rater reliability for nominal data - which
  coefficients and confidence intervals are appropriate?”, BMC Medical
  Research Methodology 16 (2016), pp. 93.
