#include "combinations.h"
#include <RcppArmadillo.h>
// [[Rcpp::depends(RcppArmadillo)]]

typedef std::function<double(arma::vec)> fun;

// [[Rcpp::export]]
double quadratic(arma::vec x) {
  arma::vec means = arma::ones(x.size()) * arma::mean(x);
  return pow(arma::norm(x - means, 2), 2.0);
}

// [[Rcpp::export]]
double absolute(arma::vec x) {
  arma::vec medians = arma::ones(x.size()) * arma::median(x);
  return arma::norm(x - medians, 1);
}

// [[Rcpp::export]]
double nominal(arma::vec x) {
  std::map<unsigned long, unsigned long> counter;
  unsigned long current_max = 0;

  for(const auto& elem : x) {
    counter.emplace(elem, 0);
    counter[elem] += 1;
    if (counter[elem] > current_max) {
      current_max = counter[elem];
    }
  }
  return x.size() - current_max;
}

// [[Rcpp::export]]
double hubert(arma::vec x) {
  return arma::all(x == x(0)) ? 0 : 1;
}

double product(arma::vec y, std::vector<unsigned int> u) {
  double out = 1;
  for (auto &i : u) {
    out *= y[i];
  }
  return out;
}

arma::vec y(std::vector<unsigned int> u, unsigned int row, unsigned int col, arma::mat x, arma::vec uniques, unsigned int g) {
  arma::vec result(g);
  result[0] = x(row, col);
  for (unsigned int i = 1; i < g; i++) {
    result[i] = uniques[u[i - 1]];
  }
  return result;
}

class mu {
  arma::mat ratings;
  fun disagreement;
  unsigned int g; // Arity of disagreement function.
  long count = 0;
  arma::vec result = arma::zeros(ratings.n_rows);

public:
  explicit mu(arma::mat x, fun d, unsigned int g) : ratings(x), disagreement(d), g(g) {
  }

  template <class It>
  arma::mat into_vec(arma::rowvec y, It begin, It end) {
    arma::vec z(g);
    unsigned int i = 0;
    for(; begin != end; ++begin) {
      z(i) = y[*begin];
      i++;
    }
    return z;
  }

  template <class It>
  bool operator()(It first, It last) {
    for(unsigned int i = 0; i < ratings.n_rows; i++) {
      arma::vec y = into_vec(ratings.row(i), first, last);
      result(i) += disagreement(y);
    }
    count++;
    return false;
  }

  operator arma::vec() const {
    return result / count;
  }
};

// Calculates the `muds` in a disagreement study.
//
// @param x A matrix of ratings.
// @param g The arity of the disagreement function.
// @param disagreement The disagreement function used.
// @return Vector of `muds`.

arma::vec muds(arma::mat x, unsigned int g, fun disagreement) {
  std::vector<int> v(x.n_cols);
  arma::vec return_vector(x.n_rows);
  std::iota(v.begin(), v.end(), 0);
  return for_each_combination(
    v.begin(),
    v.begin() + g,
    v.end(),
    mu(x, disagreement, g));
}

// Calculates the `mus` for the Cohen-type agreement.
//
// @param x Matrix of ratings.
// @param uniques Vector of unique values.
// @param pk Vector of probabilities.
// @param g The arity of the disagreement function.
// @param disagreement The disagreement function used.
// @return Vector of `muds`.

arma::vec mus_cohen(arma::mat x, arma::vec uniques, arma::mat pk, unsigned int g, fun disagreement, arma::mat r_combs) {

  arma::vec result(x.n_rows);
  unsigned int n = uniques.size();
  std::vector<unsigned int> u(g - 1);

  for(long k = 0; k < pow(n, g - 1); ++k ) {
    lldiv_t q{k, 0};
    for(long j = g - 2; 0 <= j; --j) {
      q = div(q.quot, n);
      u[j] = q.rem;
    }

    for(unsigned int j = 0; j < r_combs.n_rows; j++) {
      for(unsigned int i = 0; i < r_combs.n_cols; i++) {

        unsigned int col = r_combs(j, i);
        double prob_use = 1;
        unsigned int rater_index = 0;

        for(unsigned int l = 0; l < g - 1; l++) {
          if(rater_index == i) rater_index++;
          prob_use *= pk(u[l], r_combs(j, rater_index));
          rater_index++;
        }

        for(unsigned int row = 0; row < x.n_rows; row++) {
          result(row) += disagreement(y(u, row, col, x, uniques, g)) * prob_use;
        }

      }
    }

  }
  return result / (r_combs.n_rows * r_combs.n_cols);
}


// Calculates the `mus` for the Fleiss-type agreement.
//
// @param x Matrix of ratings.
// @param uniques Vector of unique values.
// @param pk Vector of probabilities.
// @param g The arity of the disagreement function.
// @param disagreement The disagreement function used.
// @return Vector of `muds`.

arma::vec mus_fleiss(arma::mat x, arma::vec uniques, arma::mat pk, unsigned int g, fun disagreement) {
  arma::vec result(x.n_rows);

  unsigned int n = uniques.size();
  std::vector<unsigned int> u(g - 1);

  // This loop iterates over n ** (g - 1), where n is the number of unique
  // elements in the data set.

  // Thus u equals a row from `i_combs` in the R file.

  for(long k = 0; k < pow(n, g - 1); ++k ) {
    lldiv_t q{k, 0};

    for(long j = g - 2; 0 <= j; --j) {
      q = div(q.quot, n);
      u[j] = q.rem;
    }

    // The vector u contains the indices of the unique values we care about
    // in this iteration.

    // Iteration over the rows are necessary.
    for(unsigned int row = 0; row < x.n_rows; row++) {
      for(unsigned int col = 0; col < x.n_cols; col++) {
        double prob = product(pk.col(col), u);

        // Making this function into a class would allow us to make this call
        // simpler, such as y(u, row, col), as uniues, x, and g are already
        // implicit.

        result(row) += disagreement(y(u, row, col, x, uniques, g)) * prob;
      }
    }
  }
  return result / x.n_cols;
}



// [[Rcpp::export]]
arma::vec mus_fleiss_cpp(arma::mat x, arma::vec uniques, arma::mat pk, unsigned int g, unsigned int disagreement) {
  switch(disagreement) {
  case 0:
    return mus_fleiss(x, uniques, pk, g, nominal);
  case 1:
    return mus_fleiss(x, uniques, pk, g, absolute);
  case 2:
    return mus_fleiss(x, uniques, pk, g, quadratic);
  case 3:
    return mus_fleiss(x, uniques, pk, g, hubert);
  }
  return mus_fleiss(x, uniques, pk, g, hubert);
}

// [[Rcpp::export]]
arma::vec mus_cohen_cpp(arma::mat x, arma::vec uniques, arma::mat pk, unsigned int g, unsigned int disagreement, arma::mat r_combs) {
  switch(disagreement) {
  case 0:
    return mus_cohen(x, uniques, pk, g, nominal, r_combs);
  case 1:
    return mus_cohen(x, uniques, pk, g, absolute, r_combs);
  case 2:
    return mus_cohen(x, uniques, pk, g, quadratic, r_combs);
  case 3:
    return mus_cohen(x, uniques, pk, g, hubert, r_combs);
  }
  return mus_cohen(x, uniques, pk, g, hubert, r_combs);
}

// [[Rcpp::export]]
arma::vec muds_simple_cpp(arma::mat x, unsigned int g, unsigned int disagreement) {
  switch(disagreement) {
  case 0:
    return muds(x, g, nominal);
  case 1:
    return muds(x, g, absolute);
  case 2:
    return muds(x, g, quadratic);
  case 3:
    return muds(x, g, hubert);
  }
  return muds(x, g, hubert);
}

// Get estimate and standard deviation.
//
// @param muds,mus,g Used in the calculation of kappas and their sd.
// @return Vector of estimates and standard deviations.
// [[Rcpp::export]]
arma::rowvec2 est_and_sd(arma::vec muds, arma::vec mus, int g) {
  double d_hat = arma::mean(muds);
  double c_hat = arma::mean(mus);
  double c_hat2 = c_hat * c_hat;
  arma::mat cov = arma::cov(muds, mus);
  arma::vec ss = {arma::var(muds), g*cov(0, 0), g*g*arma::var(mus)};
  arma::vec ps = {1, -2*d_hat / c_hat, d_hat * d_hat / c_hat2};
  ps = ps / c_hat2;

  return {1 - d_hat / c_hat, arma::dot(ps, ss)};
}

